import logging
import re

from bs4 import BeautifulSoup
import requests
from urllib.parse import urljoin
import urllib3

from .utils import print_atomic
from .config import Config

logger = logging.getLogger(__name__)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)  # seen due to web module requests.get(verify=False)


class Web:
    """
    :return: self.get = [http status | error, shortened text, original html, redirects, x-frame-options, csp, form_names]
    """
    cache = {}
    store_html = True
    store_text = True
    headers = {}

    @classmethod
    def init(cls, store_text=True, store_html=True):
        cls.store_text = store_text
        cls.store_html = store_html
        if Config.get("user_agent", "FIELDS"):
            cls.headers = {"User-Agent": Config.get("user_agent", "FIELDS")}

    def __init__(self, url):
        if url in self.cache:
            self.get = self.cache[url]
            return
        redirects = []
        current_url = url
        while True:
            try:
                logger.debug(f"Scrapping connection to {current_url}")
                response = requests.get(current_url, timeout=Config.get("web_timeout", "FIELDS", get=int),
                                        headers=self.headers,
                                        allow_redirects=False, verify=False)
            except IOError as e:
                if isinstance(e, requests.exceptions.HTTPError):
                    s = 0
                elif isinstance(e, requests.exceptions.ConnectionError):
                    s = -1
                elif isinstance(e, requests.exceptions.RequestException):
                    s = -2
                elif isinstance(e, requests.exceptions.Timeout):
                    s = -3
                else:
                    s = e
                self.cache[url] = self.get = str(s), None, None, redirects, None, None, None
                print_atomic(f"Scrapping {url} failed: {e}")
                break
            if response.headers.get("Location") and len(redirects) < 10:
                current_url = urljoin(current_url, response.headers.get("Location"))
                redirects.append(current_url)
                continue
            else:
                response.encoding = response.apparent_encoding  # https://stackoverflow.com/a/52615216/2036148
                if self.store_text:
                    soup = BeautifulSoup(response.text, features="html.parser")
                    # check redirect
                    res = soup.select("meta[http-equiv=refresh i]")
                    if res:
                        wait, txt = res[0].attrs["content"].split(";")
                        m = re.search(r"http[^\"'\s]*", txt)
                        if m and len(redirects) < 10:
                            current_url = m.group(0)
                            redirects.append(current_url)
                            continue
                    # prepare content to be shortened
                    [s.extract() for s in
                     soup(["style", "script", "head"])]  # remove tags with low probability of content
                    text = re.sub(r'\n\s*\n', '\n', soup.text)  # reduce multiple new lines to singles
                    text = re.sub(r'[^\S\r\n][^\S\r\n]*[^\S\r\n]', ' ',
                                  text)  # reduce multiple spaces (not new lines) to singles

                    # if the form tag like <input> or <select> has no attribute "name", print out its tag name and value or options
                    def get_info(el):
                        """ This element has no "name" attribute """
                        n = el.name
                        r = [n]
                        if n == "select":
                            for opt in el.find_all("option"):
                                r.append(opt.attrs.get("value", "") + ":" + opt.text)
                        else:
                            r.append(el.attrs.get("value", ""))
                        return " ".join(r)

                    form_names = [s.attrs.get("name", get_info(s)) for s in soup(("input", "select", "textarea"))]
                else:
                    form_names = None
                    text = ""
                # for res in response.history[1:]:
                #     redirects += f"REDIRECT {res.status_code} → {res.url}\n" + text
                #     redirects.append(res.url)

                print_atomic(f"Scrapped {url} ({len(response.text)} bytes)")
                self.cache[
                    url] = self.get = response.status_code, text.strip(), response.text if self.store_html else None, \
                    redirects, \
                    response.headers.get('X-Frame-Options', None), \
                    response.headers.get('Content-Security-Policy', None), \
                    form_names
                break
                # if current_url == url:
                #     break
                # url = current_url



import os
from pathlib import Path
from unittest import TestCase

from tests.shared import GIF_CSV, PROJECT_DIR, SHEET_CSV, TESTDATA_DIR, Convey, p


class TestSending(TestCase):
    def test_dynamic_template(self):
        convey = Convey("--output", "False", filename=SHEET_CSV)
        cmd = """--field code,3,'x="example@example.com" if "example.com" in x else x+"@example.com"'""" \
              " --split code --send-test {mail} 'email_template.eml' --headless"
        
        os.chdir(PROJECT_DIR / TESTDATA_DIR)

        lines = convey(cmd.format(mail="example@example.com"))
        self.assertIn('Subject: My cool dynamic template demonstrating a long amount of lines!', lines)
        self.assertIn('We send you lots of colours: red, green, yellow.', lines)
        self.assertIn('foo,green,first.example.com,example@example.com', lines)
        # attachment must not be present because we called attachment() in the template
        self.assertNotIn('Attachment', lines[0])

        lines = convey(cmd.format(mail="wikipedia.com@example.com"))
        self.assertIn('Subject: My cool dynamic template demonstrating a short amount of lines!', lines)
        self.assertIn('We send you single colour: orange.', lines)
        self.assertIn('Attachment sheet.csv (text/csv):', lines[0])

        cmd += " --header"  # we force first row to be a header
        lines = convey(cmd.format(mail="wikipedia.com@example.com"))
        # even though there is header in the file, we should still get single value
        self.assertIn('We send you single colour: orange.', lines)

    # XX we should test body, subject, references flag
    # def test_body_flag(self):
    #     convey = Convey(FILTER_FILE)
    #     cmd = """--body "body text" """

    def test_send(self):
        BLACK = "Attachment black.gif (image/gif)"
        WHITE = "Attachment white.gif (image/gif)"
        COLOURS = "Attachment gif.csv (text/csv)"
        convey = Convey("--output", "False", filename=GIF_CSV)
        cmd_pattern = "-t abusemail,path --split abusemail --send-test {mail} 'bare_template.eml' "

        cmd = cmd_pattern + "--attach-files False --attach-paths-from-path-column True"

        os.chdir(PROJECT_DIR / TESTDATA_DIR)

        # Single image is attached
        lines = convey(cmd.format(mail="john@example.com"))
        self.assertIn(BLACK, lines[0])
        lines = convey(cmd.format(mail="mary@example.com"))
        self.assertIn(WHITE, lines[0])

        # Two images are attached
        lines = convey(cmd.format(mail="jack@example.com"))
        self.assertIn(BLACK, lines[0])
        self.assertIn(WHITE, lines[4])

        # Image cannot be attached
        lines = convey(cmd.format(mail="hyacint@example.com"))
        self.assertIn(
            "Convey crashed at For security reasons, path must be readable to others: red-permission.gif", lines[0])

        # Flags controlling attachments work
        lines = convey(cmd_pattern.format(mail="john@example.com") +
                       "--attach-files True --attach-paths-from-path-column False")
        self.assertIn(COLOURS, lines[0])
        lines = convey(cmd_pattern.format(mail="john@example.com") +
                       "--attach-files True --attach-paths-from-path-column True")
        self.assertIn(COLOURS, lines[0])
        self.assertIn(BLACK, lines[1])
        lines = convey(cmd_pattern.format(mail="john@example.com") +
                       "--attach-files False --attach-paths-from-path-column False")
        self.assertNotIn(COLOURS, lines[0])
        self.assertNotIn(BLACK, lines[1])



from unittest import TestCase

from tests.shared import SHEET_CSV, Convey


class TestDialect(TestCase):
    def test_dialect(self):
        convey = Convey(SHEET_CSV)
        self.assertIn("foo|red|second.example.com", convey("--delimiter-output '|'"))
        self.assertIn("foo|red|second.example.com", convey("--delimiter-output '|' --header-output false"))
        self.assertNotIn("foo|red|second.example.com", convey("--delimiter-output '|' --header-output false --header"))



from tests.shared import SHEET_CSV, Convey, TestAbstract


class TestColumns(TestAbstract):
    def test_column_selection(self):
        """ Allows specify index (even negative) """

        self.check("com", "-f tld,-1", "one.com")
        self.check('"one.com","com"', "-f tld,-1 -C", "one.com")
        self.check("com", "-f tld,1", "one.com")
        self.check("Column ID 2 does not exist. We have these so far: one.com", "-f tld,2", "one.com")
        self.check('two.cz,one.com,com', "-f tld,2 -C", "two.cz,one.com")

        c = Convey(filename=SHEET_CSV)
        self.assertEqual('foo,green,first.example.com,com', c("-f tld,-1")[1])
        self.assertEqual('foo,green,first.example.com,com', c("-f tld,3")[1])
        self.assertEqual('foo,green,first.example.com,com,comA', c("-f tld,-1 -f code,-1,'x+=\"A\"'")[1])
        self.assertEqual('foo,green,first.example.com,com,first.example.comA', c("-f tld,-1 -f code,-2,'x+=\"A\"'")[1])

    def test_split(self):
        lines = Convey()("--split email", "one@example.com\nsecond@example.com")
        [self.assertIn(s, lines) for s in
         ('* Saved to second@example.com', '"second@example.com"', '* Saved to one@example.com', '"one@example.com"')]



from base64 import b64encode
from datetime import datetime

from tests.shared import HELLO_B64, Convey, TestAbstract

convey = Convey()

class TestFields(TestAbstract):

    def test_base64_detection(self):
        """Base64 detection should work if majority of the possibly decoded characters are not gibberish."""
        self.check("base64", "--single-detect", HELLO_B64)
        self.check("", "--single-detect", "ahojahoj")

    def test_base64_charset(self):
        """Base64 detection should work even if encoded with another charset"""
        s = "Žluťoučký kůň pěl ďábelské ódy."
        encoded = b64encode(s.encode("iso-8859-2"))
        convey = Convey()
        self.assertIn(s, convey("-f charset,,,iso-8859-2", text=encoded.decode("utf-8")))

    def test_base64_disambiguation(self):
        """Base64 must not mix up with I.E. hostname"""
        c = Convey("--single-detect")
        self.assertIn("hostname", c("example.com"))  # hostname must not be confounded with base64
        self.assertFalse(c("base"))  # 'base' is plaintext

        c = Convey("--single-query")
        self.assertIn("m«", c("base -t base64"))  # 'base' can be base64 if explicitly told

    def test_phone_detection(self):
        """Various phone formats must pass."""
        c = Convey("--single-detect")
        self.assertIn("timestamp", c("2020-02-29"))  # date value must not be confused with the phone regex
        for phone in ("+420123456789", "+1-541-754-3010", "1-541-754-3010", "001-541-754-3010", "+49-89-636-4801"):
            self.assertIn("phone", c(phone), phone)

    def test_pint(self):
        """Test unit conversion works"""
        c = Convey()
        self.assertIn("unit", c("--single-detect", text="1 kg"))
        self.assertIn("2.6792288807189983 troy_pound", c("-f unit[troy_pound]", text="1 kg"))

    def test_wrong_url(self):
        c = Convey()
        self.assertEqual("http://example.com", c("-f url", text="hXXp://example.com")[0])
        self.assertEqual("https://an.eXAmple.com", c("-f url", text="hxxps://an[.]eXAmple[.]com")[0])
        self.assertEqual("http://185.33.144.243/main_content/", c("-f url", text="hxxp://185.33.144[.]243/main_content/")[0])
        self.assertEqual("http://80.211.218.7/fb/", c("-f url", text="80.211.218.7/fb/")[0])

    def test_hostname(self):
        self.assertIn("hostname", convey("--single-detect", text="_spf.google.com"))

    def test_timestamp(self):
        self.assertIn("timestamp", convey("--single-detect", text="26. 03. 1999"))
        time = datetime.now()
        self.assertIn("timestamp", convey("--single-detect", text=str(time)))
        self.assertIn("timestamp", convey("--single-detect", text=str(int(time.timestamp()))))

        # as of Python3.7 use:
        # distant_future = datetime.fromisoformat("3000-01-01")  # it is less probable distant dates are dates
        distant_future = datetime.fromtimestamp(32503680000.0)

        self.assertIn("timestamp", convey("--single-detect", text=str(distant_future)))
        self.assertIn("phone", convey("--single-detect", text=str(int(distant_future.timestamp()))))
        # there is no path to datetype date from a phone
        self.assertIn("No suitable column found for field 'date'", convey("-S -f date", text=str(distant_future.timestamp())))
        # however, is is possible to get a date if specified
        self.assertIn("3000-01-01", convey("-t timestamp -f date", text=str(int(distant_future.timestamp()))))
        # works for float numbers too
        self.assertIn("3000-01-01", convey("-t timestamp -f date", text=str(distant_future.timestamp())))

        # short number is not considered a timestamp from the beginning of the Unix epoch (1970)
        self.assertEqual([], convey("--single-detect", text="12345"))



import os
from pathlib import Path
from shared import (
    CONSUMPTION,
    COMBINED_LIST_METHOD,
    COMBINED_SHEET_PERSON,
    GIF_CSV,
    PERSON_CSV,
    PERSON_GIF_CSV,
    PERSON_HEADER_CSV,
    SHEET_CSV,
    SHEET_HEADER_CSV,
    SHEET_HEADER_ITSELF_CSV,
    SHEET_HEADER_PERSON_CSV,
    SHEET_PERSON_CSV,
    TestAbstract,
    Convey,
    p,
)


class TestAction(TestAbstract):
    def test_aggregate(self):
        self.check(
            ["sum(price)", "972.0"], f"--aggregate price,sum", filename=CONSUMPTION
        )
        self.check(
            ["category,sum(price)", "total,972.0", "kettle,602.0", "bulb,370.0"],
            f"--aggregate price,sum,category",
            filename=CONSUMPTION,
        )
        self.check(
            [
                "category,sum(price),avg(consumption)",
                "total,972.0,41.0",
                "kettle,602.0,75.0",
                "bulb,370.0,18.33",
            ],
            f"--aggregate price,sum,consumption,avg,category",
            filename=CONSUMPTION,
        )
        self.check(
            [
                "category,sum(price),list(price)",
                "total,972.0,(all)",
                '''kettle,602.0,"['250', '352']"''',
                '''bulb,370.0,"['100', '150', '120']"''',
            ],
            f"--aggregate price,sum,price,list,category",
            filename=CONSUMPTION,
        )

        # XX this will correctly split the files,
        # however, the output is poor and for a reason not readable by the check.
        # self.check(['','Split location: bulb','','Split location: kettle'],
        #            "--agg price,sum --split category", filename=CONSUMPTION)
        # Until then, following substitution is used to generate the files at least
        Convey(filename=CONSUMPTION)("--agg price,sum --split category")

        # Check the contents of the files that just have been split
        check1 = False
        check2 = False
        for f in Path().rglob("consumption.csv_convey*/*"):
            if f.name == "kettle" and f.read_text() == "sum(price)\n602.0\n":
                check1 = True
            if f.name == "bulb" and f.read_text() == "sum(price)\n370.0\n":
                check2 = True
        self.assertTrue(check1)
        self.assertTrue(check2)

    def test_aggregate_group_col(self):
        # group by a column without any additional info means counting
        self.check(
            [
                "price,count(price)",
                "total,5",
                "100,1",
                "150,1",
                "250,1",
                "352,1",
                "120,1",
            ],
            f"-a price",
            filename=CONSUMPTION,
        )

        # group by the same column works
        self.check(
            [
                "price,sum(price)",
                "total,972.0",
                "352,352.0",
                "250,250.0",
                "150,150.0",
                "120,120.0",
                "100,100.0",
            ],
            f"--aggregate price,sum,price",
            filename=CONSUMPTION,
        )

        self.check(
            [
                "price,count(price)",
                "total,5",
                "100,1",
                "150,1",
                "250,1",
                "352,1",
                "120,1",
            ],
            f"--aggregate price,count,price",
            filename=CONSUMPTION,
        )

        # group by a different column when counting does not make sense
        msg = "ERROR:convey.action_controller:Count column 'price' must be the same as the grouping column 'consumption'."
        with self.assertLogs(level="WARNING") as cm:
            self.check("", f"--aggregate price,count,consumption", filename=CONSUMPTION)
            self.assertEqual([msg], cm.output)

    def test_merge(self):
        # merging generally works
        self.check(
            COMBINED_SHEET_PERSON, f"--merge {PERSON_CSV},2,1", filename=SHEET_CSV
        )

        # rows can be duplicated due to other fields
        self.check(
            COMBINED_LIST_METHOD,
            f"--merge {PERSON_CSV},2,1 -f external,1," + str(p("external_pick_base.py")) + ",list_method",
            filename=SHEET_CSV,
        )

        # merging file with header and with a missing value
        self.check(
            SHEET_PERSON_CSV, f"--merge {PERSON_HEADER_CSV},2,1", filename=SHEET_CSV
        )

        # merge on a column type
        self.check(
            PERSON_GIF_CSV, f"--merge {GIF_CSV},email,email", filename=PERSON_CSV
        )

        # merge by a column number
        self.check(PERSON_GIF_CSV, f"--merge {GIF_CSV},email,1", filename=PERSON_CSV)

        # invalid column definition
        msg = "ERROR:convey.identifier:Cannot identify COLUMN invalid, put there an exact column name, its type, the numerical order starting with 1, or with -1."
        with self.assertLogs(level="WARNING") as cm:
            self.check("", f"--merge {GIF_CSV},email,invalid", filename=PERSON_CSV)
            self.assertEqual([msg], cm.output)
        # merging a file with itself
        self.check(
            SHEET_HEADER_ITSELF_CSV,
            f"--merge {SHEET_HEADER_CSV},4,2",
            filename=SHEET_HEADER_CSV,
        )

        # only local file has header; different dialects
        self.check(
            SHEET_HEADER_PERSON_CSV,
            f"--merge {PERSON_CSV},2,1",
            filename=SHEET_HEADER_CSV,
        )

    def test_compute_from_merge(self):
        """Computing a new column from another file currenlty being merged was not implemented."""
        self.check(
            "Column ID 6 does not exist. We have these so far: foo, red, second.example.com",
            f"--merge {PERSON_CSV},2,1 -f base64,6",
            filename=SHEET_CSV,
        )



from typing import Callable, Generator, Union

AggregateMethod = Callable
"A method of the Aggregate class, ex: Aggregate.sum"

class Aggregate:

    def __init__(self, factory: AggregateMethod):
        self.count: Union[int, str] = 0
        "count or description"
        self.generator: Generator = factory()
        next(self.generator)

        self.is_roundable = factory in (Aggregate.sum, Aggregate.avg)
        self.is_too_broad = factory is self.list

    def get(self):
        return round(self.count, 2) if self.is_roundable else self.count

    @classmethod
    def all(cls):
        return [cls.avg, cls.sum, cls.count, cls.min, cls.max, cls.list, cls.set]

    @staticmethod
    def avg():
        count = 0
        res = float((yield))
        while True:
            count += 1
            res += float((yield res / count))

    @staticmethod
    def sum():
        res = 0
        while True:
            res += float((yield res))

    @staticmethod
    def count():
        res = 0
        while True:
            yield res
            res += 1

    @staticmethod
    def min():
        v = yield
        while True:
            v = min((yield v), v)

    @staticmethod
    def max():
        v = yield
        while True:
            v = max((yield v), v)

    @staticmethod
    def list():
        l = []
        while True:
            l.append((yield l))

    @staticmethod
    def set():
        s = set()
        while True:
            s.add((yield s))

    # XX If we would like to serialize a function and this is not possible, we can serialize it ourselves that way:
    # @staticmethod
    # def avg():
    #     count = 0
    #     try:
    #         res = float((yield))
    #     except LoadFromSerialization:
    #         count, res = yield
    #     try:
    #         while True:
    #             count += 1
    #             res += float((yield res / count))
    #     except StopIteration: -> serialization request
    #         yield count, res


aggregate_functions = [f.__name__ for f in Aggregate.all()]
aggregate_functions_str = "".join("\n* " + f for f in aggregate_functions)


from unittest import TestCase

from tests.shared import SHEET_CSV, Convey


class TestFilter(TestCase):
    def test_filter(self):
        convey = Convey(SHEET_CSV)
        self.assertEqual(3, len(convey("--include-filter 1,foo")))
        self.assertEqual(2, len(convey("--exclude-filter 1,foo")))
        self.assertEqual(2, len(convey("--unique 1")))

    def test_post_filter(self):
        """Filter after a field was generated."""
        convey = Convey(SHEET_CSV)
        self.assertEqual(3, len(convey("--field base64,1 --include-filter base64,Zm9v")))
        self.assertEqual(2, len(convey("--field base64,1 --exclude-filter base64,Zm9v")))
        self.assertEqual(2, len(convey("--field base64,1 --unique base64")))



from tests.shared import GIF_CSV, PERSON_CSV, PERSON_GIF_CSV, TestAbstract


class TestInternal(TestAbstract):
    def test_similar_fields(self):
        """Recommending of the similar columns"""
        c1 = self.check(None, f"--merge {PERSON_CSV},2,1", filename=GIF_CSV)
        parser1A, parser1B = c1.parser, c1.parser.settings["merge"][0].remote_parser
        fields1A = c1.parser.fields
        fields1B = parser1B.fields
        self.assertListEqual([fields1A[0]], parser1A.get_similar(fields1B))
        self.assertListEqual([fields1B[0]], parser1B.get_similar(fields1A))
        self.assertListEqual([fields1B[0]], parser1B.get_similar(fields1A[0]))
        self.assertListEqual([], parser1B.get_similar(fields1A[1]))

        c2 = self.check(None, f"--merge {PERSON_GIF_CSV},2,1", filename=GIF_CSV)
        parser2A, parser2B = c2.parser, c2.parser.settings["merge"][0].remote_parser
        fields2A = c2.parser.fields
        fields2B = parser2B.fields
        self.assertListEqual([fields2A[0], fields2A[1]], parser2A.get_similar(fields2B))
        self.assertListEqual([fields2A[0]], parser2A.get_similar(fields2B[0]))
        self.assertListEqual([], parser2A.get_similar(fields2B[1]))
        self.assertListEqual([fields2A[1]], parser2A.get_similar(fields2B[3]))
        self.assertListEqual([fields2B[0], fields2B[3]], parser2B.get_similar(fields2A))



from tests.shared import SHEET_CSV, SHEET_DUPLICATED_CSV, Convey, TestAbstract, p


class TestExternals(TestAbstract):

    def test_list_in_result(self):
        """The method returns a list while CSV processing."""
        self.check(SHEET_DUPLICATED_CSV, "-f external," + str(p("external_pick_base.py")) + ",list_method", filename=SHEET_CSV)

    def test_bare_method(self):
        # single value converted by the external
        self.check("-foo-", "--field external," + str(p("external_pick_base.py")) + ",dumb_method --input 'foo'")

        # by default, first column is used
        self.check("foo,bar,-foo-", "--field external," + str(p("external_pick_base.py")) + ",dumb_method -C --input 'foo,bar'")

        # specify 2nd column is the source for the external field
        self.check("foo,bar,-bar-", "--field external,2,plaintext," + str(p("external_pick_base.py")) + ",dumb_method -C --input 'foo,bar'")

    def test_pick_input(self):
        convey = Convey()
        lines = convey("--field external," + str(p("external_pick_base.py")) + ",time_format --input '2016-08-08 12:00'")
        self.assertEqual(lines, ["12:00"])

    def test_pick_method(self):
        convey = Convey()
        # method "all" (default one) is used and "1" passes
        self.assertEqual(["1"], convey("--field external," + str(p("external_pick_base.py")) + ",PickMethodTest --input '1'"))
        # method "filtered" is used and "a" passes
        self.assertEqual(["a"], convey("--field external," + str(p("external_pick_base.py")) + ",PickMethodTest,filtered --input 'a'"))
        # method "filtered" is used which excludes "1"
        self.assertNotEqual(["1"], convey("--field external," + str(p("external_pick_base.py")) + ",PickMethodTest,filtered --input '1'"))

        # XX does not work, error when resolving path Unit -> Plaintext -> External. Identifier.get_methods_from
        # `lambda_ = lambda_.get_lambda(custom.pop(0) if custom is not None else None)`
        #       -> get_module_from_path(custom[0], ...) does not contain a path
        # self.assertEqual(["a"], convey("--field external,external_pick_base.py,PickMethodTest --input 'mm'"))



from base64 import b64encode
from pathlib import Path
import shutil
from tempfile import TemporaryDirectory
from tests.shared import COMBINED_SHEET_PERSON, HELLO_B64, PERSON_ODS, PERSON_XLS, PERSON_XLSX, SHEET_CSV, Convey, TestAbstract


class TestLaunching(TestAbstract):
    def test_piping_in(self):
        convey = Convey()
        # just string specified, nothing else
        lines = convey(piped_text="3 kg")
        self.assertTrue(len(lines) == 1)
        self.assertTrue("'1.806642228624337e+27 dalton'" in lines[0])

        # field base64 specified
        self.assertListEqual([HELLO_B64], convey("-f base64", piped_text="hello"))
        self.assertListEqual(["hello"], convey(piped_text=HELLO_B64))
        self.assertListEqual([], convey(piped_text="hello"))

    def test_single_query_processing(self):
        """Stable --single-query parsing"""

        # Single field containing a comma, still must be fully converted (comma must not be mistaken for a CSV delimiter)
        self.check("aGVsbG8sIGhlbGxv", "-f base64", "hello, hello", debug=True)
        self.check("V=C3=A1=C5=BEen=C3=A1 Ad=C3=A9lo, ra=C4=8Dte vstoupit", "-f quoted_printable", "Vážená Adélo, račte vstoupit")
        self.check([], "", "hello, hello")

        # multiline base64 string
        multiline = "ahoj\nahoj"
        multiline_64 = b64encode(multiline.encode()).decode()
        word = "ahoj"
        word_64 = b64encode(word.encode()).decode()

        self.check(multiline_64, "--single-query -f base64", multiline)
        self.check([f'"{word}","{word_64}"'] * 2, "-f base64", multiline)

    def test_conversion(self):
        lines = ["john@example.com", "mary@example.com", "hyacint@example.com"]
        with TemporaryDirectory() as temp:
            for pattern in (PERSON_XLS, PERSON_XLSX, PERSON_ODS):
                f = Path(temp, pattern.name)
                f_converted = Path(temp, pattern.name + ".csv")
                # XX as of Python3.8, use this line: shutil.copy(pattern, f)
                shutil.copy(str(pattern), str(f))  # move to temp dir to not pollute the tests folder

                self.assertFalse(f_converted.exists())
                self.check(lines, f"-s 1", filename=f)
                self.assertTrue(f_converted.exists())
                # try again, as the file now exists
                self.check(lines, f"-s 1", filename=f)

                # clean the converted file up and use it not as the main file but
                # as a secondary Wrapper – in a merge action
                f_converted.unlink()
                self.check(COMBINED_SHEET_PERSON, f"--merge {f},2,1", filename=SHEET_CSV)
                self.assertTrue(f_converted.exists())
                self.check(COMBINED_SHEET_PERSON, f"--merge {f},2,1", filename=SHEET_CSV)



import csv
import re
from difflib import SequenceMatcher
import logging
from typing import List, Optional
from pathlib import Path
from sys import exit

from dialog import Dialog
from prompt_toolkit.shortcuts import clear

from .action import AggregateAction, MergeAction
from .aggregate import Aggregate, AggregateMethod, aggregate_functions_str, aggregate_functions
from .config import Config
from .decorators import PickBase, PickMethod, PickInput
from .dialogue import Cancelled, Menu, choose_file, csv_split, hit_any_key, pick_option, ask, is_yes
from .field import Field
from .parser import Parser
from .types import Types, TypeGroup, types, Type, graph, methods, get_module_from_path
from .wizzard import Preview
from .wrapper import Wrapper

logger = logging.getLogger(__name__)


class ActionController:
    def __init__(self, parser: Parser, reprocess=False):
        self.parser = parser
        self.reprocess = reprocess

    def add_column(self):
        self.select_col("New column", only_computables=True, add=True)
        self.parser.is_processable = True

    def add_aggregation(self, fn_name: str = None, column_task: Optional[str] = None, group: Optional[Field] = None, grouping_probably_wanted=True, exit_on_fail=False):
        # choose what column we want

        fn = self.get_aggregation_fn(fn_name, exit_on_fail)
        if column_task:
            field: Field = self.parser.fields[self.parser.identifier.get_column_i(column_task, "to be aggregated with")]
        else:
            field = self.select_col("aggregation")

        sett = self.parser.settings["aggregate"]
        if sett:
            group_old, fns = sett.group_by, sett.actions
        else:
            group_old, fns = None, []

        group = self.assure_aggregation_group_by(fn, field, group or group_old, grouping_probably_wanted, exit_on_fail)
        fns.append([fn, field])
        self.parser.settings["aggregate"] = AggregateAction(group, fns)
        self.parser.is_processable = True

    def add_dialect(self):
        # XX not ideal and mostly copies Parser.__init__
        # XX There might be a table with all the csv.dialect properties or so.
        dialect = self.parser.settings["dialect"]
        while True:
            s = "What is delimiter " + (f"(default '{dialect.delimiter}')" if dialect.delimiter else "") + ": "
            dialect.delimiter = input(s) or dialect.delimiter
            if len(dialect.delimiter) != 1:
                print("Delimiter must be a 1-character string. Invent one (like ',').")
                continue
            s = "What is quoting char " + (f"(default '{dialect.quotechar}')" if dialect.quotechar else "") + ": "
            dialect.quotechar = input(s) or dialect.quotechar
            break
        dialect.quoting = csv.QUOTE_NONE if not dialect.quotechar else csv.QUOTE_MINIMAL

        if self.parser.has_header:
            if self.parser.settings['header'] is False:
                s = f"Should we remove header? "
            else:
                s = f"Should we include header? "
            if not is_yes(s):
                self.parser.settings["header"] = not self.parser.settings["header"]

        self.parser.is_processable = True

    def add_filter(self):
        menu = Menu(title="Choose a filter")
        menu.add("Unique filter", self.add_uniquing)
        menu.add("Include filter", self.add_filtering)
        menu.add("Exclude filter", lambda: self.add_filtering(False))
        menu.sout()

    def add_merge(self, remote_path=None, remote_col_i: Optional[int] = None, local_col_i: Optional[int] = None):
        if not remote_path:
            remote_path = choose_file("What file we should merge the columns from?")
        wrapper2 = Wrapper(Path(remote_path), reprocess=self.reprocess)
        parser2 = wrapper2.parser

        # dialog user and build the link between files
        controller2 = ActionController(parser2)
        column1 = self.parser.fields[local_col_i] if local_col_i is not None else None

        if remote_col_i is not None:
            column2 = parser2.fields[remote_col_i]
        else:
            high = parser2.get_similar(column1 or self.parser.fields)
            column2 = controller2.select_col("Select remote column to be merged",
                                             include_computables=False, highlighted=high)
        if not column1:
            high = self.parser.get_similar(column2)
            column1 = self.select_col(
                f"Select local column to merge '{column2}' to", include_computables=False, highlighted=high)

        # cache remote values
        operation = MergeAction.build(wrapper2.file, parser2, column2, column1)

        # build local fields based on the remotes
        for rf in parser2.fields:
            f = Field(rf.name,
                      is_chosen=False if rf is column2 else True,
                      merged_from=rf,
                      merge_operation=operation)
            self.parser.add_field(append=f)

        # prepare the operation
        self.parser.is_processable = True
        self.parser.settings["merge"].append(operation)

    def add_new_column(self, task, add=True):
        """
        :type task: str FIELD,[COLUMN],[SOURCE_TYPE], ex: `netname 3|"IP address" ip|sourceip`
        :type add: bool Add to the result.
        """
        target_type, source_field, source_type, custom = self._add_new_column(task)
        if not self.source_new_column(target_type, add, source_field, source_type, custom):
            print("Cancelled")
            exit()
        self.parser.is_processable = True
        return target_type

    def _add_new_column(self, task):
        """ Internal analysis for `add_new_column` """
        task = csv_split(task)
        custom = []
        target_type = task[0]
        m = re.search(r"(\w*)\[([^]]*)\]", target_type)
        if m:
            target_type = m.group(1)
            custom = [m.group(2)]
        try:
            target_type = types[types.index(target_type)]  # determine FIELD by exact name
        except ValueError:
            d = {t.name: SequenceMatcher(None, task[0], t.name).ratio() for t in Types.get_computable_types()}
            rather = max(d, key=d.get)
            logger.error(f"Unknown field '{task[0]}', did not you mean '{rather}'?")
            exit()
        source_field, source_type, c = self.parser.identifier.get_fitting_source(target_type, *task[1:])
        custom = c + custom
        return target_type, source_field, source_type, custom

    def add_uniquing(self, col_i=None):
        if col_i is None:
            col_i = self.select_col("unique").col_i
        self.parser.settings["unique"].append(col_i)
        self.parser.is_processable = True

    def assure_aggregation_group_by(self, fn, field, group, grouping_probably_wanted=True, exit_on_fail=False) -> Optional[Field]:
        match (fn == Aggregate.count, group is None, grouping_probably_wanted):
            case True, True, _:
                group = field
            case False, True, True:
                # here, self.select col might return None
                group = self.select_col("group by", prepended_field=("no grouping", "aggregate whole column"))
            case True, False, _:
                if field != group:
                    logger.error(f"Count column '{field.name}' must be the same"
                                    f" as the grouping column '{group.name}'.")
                    if exit_on_fail:
                        exit()
                    else:
                        raise Cancelled
        return group

    def add_filtering(self, include=True, col_i=None, val=None):
        if col_i is None:
            col_i = self.select_col("filter").col_i
        if val is None:
            s = "" if include else "not "
            val = ask(f"What value must {s}the field have to keep the line?")
        self.parser.settings["filter"].append((include, col_i, val))
        self.parser.is_processable = True

    def add_splitting(self):
        self.parser.settings["split"] = self.select_col("splitting").col_i
        self.parser.is_processable = True

    def choose_cols(self):
        # XX possibility un/check all
        chosens = [(str(i + 1), str(f), f.is_chosen) for i, f in enumerate(self.parser.fields)]
        d = Dialog(autowidgetsize=True)
        ret, values = d.checklist("What fields should be included in the output file?", choices=chosens)
        if ret == "ok":
            for f in self.parser.fields:
                f.is_chosen = False
            for v in values:
                self.parser.fields[int(v) - 1].is_chosen = True
            self.parser.is_processable = True

    def get_aggregation_fn(self, fn_name: str = None, exit_on_fail=False) -> AggregateMethod:
        if not fn_name:
            menu = Menu("Choose aggregate function", callbacks=False, fullscreen=True)
            [menu.add(f) for f in aggregate_functions]
            option = menu.sout()
            if not option:
                raise Cancelled
            fn_name = aggregate_functions[int(option) - 1]
        fn = getattr(Aggregate, fn_name, None)
        if not fn:
            if exit_on_fail:
                logger.error(
                    f"Unknown aggregate function '{fn_name}'. Possible functions are: {aggregate_functions_str}")
                exit()
            else:
                raise Cancelled
        return fn

    def select_col(self, dialog_title="", only_computables=False, include_computables=True, add=None, prepended_field=None, highlighted: Optional[List[Field]] = None) -> Optional[Field]:
        """ Starts dialog where user has to choose a column.
            If cancelled, we return to main menu automatically.
            :type prepended_field: tuple (field_name, description) If present, this field is prepended. If chosen, you receive None.
        """
        # add existing fields
        fields = [] if only_computables else self.parser.get_fields_autodetection()

        # add computable field types
        if include_computables:
            for field in Types.get_computable_types():
                if field.from_message:
                    s = field.from_message
                else:
                    node_distance = graph.dijkstra(field, ignore_private=True)
                    s = field.group.name + " " if field.group != TypeGroup.general else ""
                    s += "from " + ", ".join([str(k) for k in node_distance][:3])
                    if len(node_distance) > 3:
                        s += "..."
                fields.append((f"new {field}...", s))

        # add special prepended field
        if prepended_field:
            fields.insert(0, prepended_field)

        # launch dialog
        col_i = pick_option(fields, dialog_title, guesses=[f.col_i for f in highlighted or ()])

        # convert returned int col_i to match an existing or new column
        if prepended_field:
            col_i -= 1
            if col_i == -1:  # we hit a prepended_field, a mere description, not a real field
                return None
        if only_computables or col_i >= len(self.parser.fields):
            target_type_i = col_i if only_computables else col_i - len(self.parser.fields)
            col_i = len(self.parser.fields)
            self.source_new_column(Types.get_computable_types()[target_type_i], add=add)

        return self.parser.fields[col_i]

    def source_new_column(self, target_type, add=None, source_field: Field = None, source_type: Type = None,
                          custom: list = None):
        """ We know what Field the new column should be of, now determine how we should extend it:
            Summarize what order has the source field and what type the source field should be considered alike.
                :type source_field: Field
                :type source_type: Type
                :type target_type: Type
                :type add: bool if the column should be added to the table; None ask
                :type custom: List
                :raise Cancelled
                :return Field
        """
        if custom is None:
            # default [] would be evaluated at the time the function is defined, multiple columns may share the same function
            custom = []
        dialog = Dialog(autowidgetsize=True)
        if not source_field or not source_type:
            print("\nWhat column we base {} on?".format(target_type))
            guesses = self.parser.identifier.get_fitting_source_i(target_type)
            source_col_i = pick_option(self.parser.get_fields_autodetection(),
                                       title="Searching source for " + str(target_type),
                                       guesses=guesses)
            source_field = self.parser.fields[source_col_i]
            source_type = self.parser.identifier.get_fitting_type(source_col_i, target_type, try_plaintext=True)
            if source_type is None:
                # ask how should be treated the column as, even it seems not valid
                # list all known methods to compute the desired new_field (e.g. for incident-contact it is: ip, hostname, ...)
                choices = [(k.name, k.description)
                           for k, _ in graph.dijkstra(target_type, ignore_private=True).items()]

                # if len(choices) == 1 and choices[0][0] == Types.plaintext:
                #     # if the only method is derivable from a plaintext, no worries that a method
                #     # converting the column type to "plaintext" is not defined; everything's plaintext
                #     source_type = Types.plaintext
                # el
                if choices:
                    s = ""
                    if self.parser.sample_parsed:
                        s = f"\n\nWhat type of value '{self.parser.sample_parsed[0][source_col_i]}' is?"
                    title = f"Choose the right method\n\nNo known method for making {target_type} from column {source_field} because the column type wasn't identified. How should I treat the column?{s}"
                    code, source_type = dialog.menu(title, choices=choices)
                    if code == "cancel":
                        raise Cancelled("... cancelled")
                    source_type = getattr(Types, source_type)
                else:
                    dialog.msgbox(
                        "No known method for making {}. Raise your usecase as an issue at {}.".format(target_type,
                                                                                                      Config.PROJECT_SITE))
                    raise Cancelled("... cancelled")
            clear()

        if not custom:
            try:
                if target_type.group == TypeGroup.custom:
                    if target_type == Types.code:
                        print("What code should be executed? Change 'x'. Ex: x += \"append\";")
                        custom = Preview(source_field, source_type).code()
                    elif target_type in [Types.reg, Types.reg_m, Types.reg_s]:
                        *custom, target_type = Preview(source_field, source_type, target_type).reg()
                    elif target_type == Types.external:  # choose a file with a needed method
                        while True:
                            path = choose_file("What .py file should be used as custom source?", dialog)
                            module = get_module_from_path(path)
                            if module:
                                # inspect the .py file, extract methods and let the user choose one
                                code, method_name = dialog.menu(f"What method should be used in the file {path}?",
                                                                choices=[(x, "") for x in dir(module) if
                                                                         not x.startswith("_")])
                                if code == "cancel":
                                    raise Cancelled("... cancelled")

                                custom = path, method_name
                                break
                            else:
                                dialog.msgbox("The file {} does not exist or is not a valid .py file.".format(path))
                    if not custom:
                        raise Cancelled("... cancelled")
            except Cancelled:
                raise Cancelled("... cancelled")
            path = graph.dijkstra(target_type, start=source_type, ignore_private=True)
            for i in range(len(path) - 1):
                m = methods[path[i], path[i + 1]]
                if isinstance(m, PickBase):
                    c = None
                    if Config.get("yes"):
                        pass
                    elif type(m) is PickMethod:
                        m: PickMethod
                        code, c = dialog.menu(f"Choose subtype", choices=m.get_options())
                        if code == "cancel":
                            raise Cancelled("... cancelled")
                    elif type(m) is PickInput:
                        m: PickInput
                        c = Preview(source_field, source_type, target_type).pick_input(m)
                    custom.insert(0, c)
        if add is None:
            if dialog.yesno("New field added: {}\n\nDo you want to include this field as a new column?".format(
                    target_type)) == "ok":
                add = True

        f = Field(target_type, is_chosen=add,
                  source_field=source_field,
                  source_type=source_type,
                  new_custom=custom)
        if f.source_field.merged_from:
            hit_any_key("Unfortunately, sourcing from columns being merged was not implemented."
                        " Ask for it on Github. And for now, merge first, then add a column.")
            raise Cancelled("Sourcing from columns being merged was not implemented")
        self.parser.settings["add"].append(f)
        self.parser.add_field(append=f)
        return f



import bdb
import importlib.util
import ipaddress
import logging
import re
import socket
from abc import ABCMeta
from base64 import b64decode, b64encode
from builtins import ZeroDivisionError
from copy import copy
from enum import IntEnum
from pathlib import Path
from quopri import decodestring, encodestring
from sys import exit
from typing import List
from urllib.parse import unquote, quote

from validate_email import validate_email

from .config import Config
from .convert import any_ip_ip, port_ip_ip, port_ip_port, reAnyIp, reFqdn, reIpWithPort, reUrl, url_hostname, url_port, wrong_url_2_url, dig, nmap
from .contacts import Contacts
from .checker import Checker
from .decorators import PickBase, PickMethod, PickInput
from .graph import Graph
from .infodicts import phone_country, address_country, country_codes
from .web import Web
from .whois import Whois

logger = logging.getLogger(__name__)

types: List["Type"] = []  # all field types
methods = {}
graph = Graph()
methods_deleted = {}


class TypeGroup(IntEnum):
    general = 1
    custom = 2
    whois = 3
    dns = 4
    nmap = 5
    web = 6

    @staticmethod
    def init():
        for module in ["whois", "web", "nmap", "dig"]:
            if Config.get(module, "FIELDS") is False:
                if module == "dig":
                    module = "dns"
                getattr(TypeGroup, module).disable()

    def disable(self):
        for start, target in copy(methods):
            if start.group is self or target.group is self:
                methods_deleted[start, target] = methods[start, target]
                del methods[start, target]
        for t in Types.get_guessable_types():
            if t.group is self:
                t.is_disabled = True


class Type:
    """
    A field type Convey is able to identify or compute
    """

    def __init__(self, name, group=TypeGroup.general, description="", usual_names=[], identify_method=None,
                 is_private=False,
                 from_message=None, usual_must_match=False):
        """
        :param name: Key name
        :param description: Help text
        :param usual_names: Names this column usually has (ex: source_ip for an IP column). List of str, lowercase, no spaces.
        :param identify_method: Lambda used to identify a value conforms this field type (return true).
            If float is returned, this should be added to the score (ex: malus when we are not sure this is a base64 string).
        :param is_private: User cannot add the field type (ex: whois, user can extend only netname which is accessed through it).
        """
        self.name = name
        self.group = group
        self.usual_names = usual_names
        self.usual_must_match = usual_must_match
        self.identify_method = identify_method
        self.description = description
        self.is_private = is_private
        self.is_disabled = False  # disabled field cannot be added (and computed)
        self.from_message = from_message
        types.append(self)
        # Type considered equal when starting computing.
        # Ex: hostname from source_ip will be computed as if from an ip.
        # So that Types.source_ip will have `computing_start = ip`
        self.computing_start: Type = self
        # Types considered equal to be computed from.
        # Ex: hostname from ip may be computed even from a source_ip.
        # So that Types.ip will have `equals = [ip, source_ip]`
        self.equals: List["Types"] = [self]
        self.is_plaintext_derivable = False

    def __getstate__(self):
        return str(self)

    def __setstate__(self, state):
        self.__dict__.update(getattr(Types, state).__dict__)

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) is str:
            return self.name == other
        return self.name == other.name

    def __lt__(self, other):
        if isinstance(other, str):
            return self.name < other
        return (self.group, self.name) < (other.group, other.name)

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Type({self.name})"

    def doc(self):
        s = self.name
        if self.description:
            s += f" ({self.description})"
        if self.usual_names:
            s += f" usual names: " + ", ".join(self.usual_names)
        return s

    def __add__(self, other):  # sometimes we might get compared to string columns names from the CSV
        if isinstance(other, str):
            return self.name + " " + other
        return self.name + " " + other.name

    def __radd__(self, other):  # sometimes we might get compared to string columns names from the CSV
        if isinstance(other, str):
            return other + " " + self.name
        return self.other + " " + self.name

    def init(self):
        """ Init self.computing_start and self.equals . """
        computing_start = [stop for start, stop in methods if start is self and methods[start, stop] is True]
        equals = [start for start, stop in methods if stop is self and methods[start, stop] is True]
        if len(computing_start) == 1:
            self.computing_start = computing_start[0]
        elif len(computing_start):
            raise ConnectionAbortedError(f"Multiple 'computing_start' types defined for {self}: {computing_start}")
        if equals:
            self.equals = equals + [self]

        # check if this is plaintext derivable
        if self != Types.plaintext:
            self.is_plaintext_derivable = bool(graph.dijkstra(self, start=Types.plaintext))

    def check_conformity(self, samples, has_header, field):
        """
        :rtype: int|False Score if the given info conforms to this type.
        """
        score = 0
        # print("Guess:", key, names, checkFn)
        # guess field type by name

        if has_header:
            s = str(field).replace(" ", "").replace("'", "").replace('"', "").lower()
            for n in self.usual_names:
                if s in n or n in s:
                    # print("HEADER match", field, self, self.usual_names)
                    score += 2 if self.usual_must_match else 1
                    break
        if not score and self.usual_must_match:
            return False
        # else:
        # guess field type by few values
        hits = 0
        for val in samples:
            sc = self.identify_method(val)
            if sc:
                hits += 1
                if isinstance(sc, float):
                    score += sc

        try:
            percent = hits / len(samples)
        except ZeroDivisionError:
            percent = 0
        if percent == 0:
            return False
        elif percent > 0.6:
            # print("Function match", field, checkFn)
            score += 1
            if percent > 0.8:
                score += 1
        return score


class Types:
    """
    Methods sourcing from private type should return a tuple, with the private type as the first
    Ex: (whois, asn): lambda x: (x, x.get[3])

    """

    @staticmethod
    def refresh():
        """ refreshes methods and import custom methods from files """
        methods.clear()
        methods.update(Types._get_methods())
        graph.clear()
        [graph.add_edge(to, from_) for to, from_ in methods if methods[to, from_] is not True]
        [t.init() for t in types]

        if Config.get("disable_external", get=bool) is True:
            return

        try:
            externals = Config.config["EXTERNAL"]
        except KeyError:
            externals = []
        for field_name in externals:
            if field_name == "external_fields":  # this is a genuine field, user did not make it
                continue
            path, method_name = Config.config["EXTERNAL"][field_name].rsplit(":")
            module = get_module_from_path(path)
            Types.import_method(module, method_name, path, name=field_name)

        for path in (x.strip() for x in Config.get("external_fields", "EXTERNAL", get=str).split(",") if x.strip()):
            # noinspection PyBroadException
            try:
                module = get_module_from_path(path)
                if module:
                    for method_name in (x for x in dir(module) if not x.startswith("_")):
                        Types.import_method(module, method_name, path)
            except bdb.BdbQuit:
                raise
            except Exception as e:
                s = f"Cannot import custom file from path: {path}"
                input(s + ". Press any key...")
                logger.warning(s)

    @staticmethod
    def import_method(module, method_name, path, name=None):
        if not name:
            name = method_name
        lambda_ = getattr(module, method_name)
        if isinstance(lambda_, ABCMeta):  # this is just an import statement of ex: PickBase
            return
        doc = lambda_.__doc__ if not isinstance(lambda_, PickBase) else lambda_.get_type_description()
        # if Config.get("disable_external", get=bool) is True:
        #     # user do not want to allow externals to be added but we have added them before argparse was parsed
        #     # so that we could inform the user in the help text of the computable fields
        #     types.pop(types.index(getattr(Types, name)))
        #     delattr(Types, name)
        # else:
        setattr(Types, name, Type(name, TypeGroup.general, doc))
        type_ = getattr(Types, name)
        methods[(Types.plaintext, type_)] = lambda_
        if lambda_ is not True:
            graph.add_edge(Types.plaintext, type_)
        type_.init()
        # Xif Config.get("disable_external", get=bool) is False:
        # Xwhen "disable_external" is None it means this method is called before argparse flags are parsed,
        # Xwe do not know yet if "disable_external" will be set to True or False
        logger.debug(f"Successfully added method {method_name} from module {path}")

    @staticmethod
    def find_type(source_type_candidate):
        source_type = None
        source_t = source_type_candidate.lower().replace(" ", "")  # make it seem like a usual field name
        possible = None
        for t in types:
            if source_t == t:  # exact field name
                source_type = t
                break
            if source_t in t.usual_names:  # usual field name
                possible = t
        else:
            if possible:
                source_type = possible
        return source_type

    @staticmethod
    def get_method(start: Type, target: Type):
        try:
            return methods[start, target]
        except KeyError:
            if (start, target) in methods_deleted:
                print(f"Disabled path at {start} – {target}. Launch --config to enable it.")
                exit()
            else:
                raise LookupError

    whois = Type("whois", TypeGroup.whois, "ask whois servers", is_private=True)
    whoisdomain = Type("whoisdomain", TypeGroup.whois, "ask whois servers for domain", is_private=True)
    web = Type("web", TypeGroup.web, "scrape web contents", is_private=True)

    external = Type("external", TypeGroup.custom, from_message="from a method in your .py file")
    code = Type("code", TypeGroup.custom, from_message="from a code you write")
    reg = Type("reg", TypeGroup.custom, from_message="from a regular expression")
    reg_s = Type("reg_s", TypeGroup.custom, from_message="substitution from a regular expression")
    reg_m = Type("reg_m", TypeGroup.custom, from_message="match from a regular expression")
    netname = Type("netname", TypeGroup.whois)
    country = Type("country", TypeGroup.whois)
    registrar_abusemail = Type("registrar_abusemail", TypeGroup.whois, "Abuse e-mail contact from whois")
    abusemail = Type("abusemail", TypeGroup.whois, "Abuse e-mail contact from whois")
    prefix = Type("prefix", TypeGroup.whois)  # XX rename to 'inetnum'? to 'range'?
    csirt_contact = Type("csirt_contact", TypeGroup.whois,
                         "E-mail address corresponding with country code, taken from your personal contacts_abroad CSV"
                         " in the format `country,abusemail`. See config.ini/contacts_abroad")
    incident_contact = Type("incident_contact", TypeGroup.whois)
    cc_contact = Type("cc_contact", TypeGroup.whois,
                      "E-mail address corresponding with the abusemail, taken from your personal contacts_cc CSV"
                      " in the format `domain,cc;cc` (mails delimited by a semicolon). See config.ini/contacts_cc."
                      " Note:  when sending an e-mail to a csirt_contact, the e-mail template fills Cc header"
                      " from the csirt_contact through contacts_cc.csv too. So you may end up with another Cc field"
                      " than in cc_contact.")
    text = Type("text", TypeGroup.web)
    http_status = Type("http_status", TypeGroup.web, "HTTP response status. If 0 or negative, request failed.")
    html = Type("html", TypeGroup.web)
    redirects = Type("redirects", TypeGroup.web)
    x_frame_options = Type("x_frame_options", TypeGroup.web)
    csp = Type("csp", TypeGroup.web)
    form_names = Type("form_names", TypeGroup.web)
    ports = Type("ports", TypeGroup.nmap, "Open ports given by nmap")
    spf = Type("spf", TypeGroup.dns)
    txt = Type("txt", TypeGroup.dns)
    a = Type("a", TypeGroup.dns)
    aaaa = Type("aaaa", TypeGroup.dns)
    ns = Type("ns", TypeGroup.dns)
    mx = Type("mx", TypeGroup.dns)
    dmarc = Type("dmarc", TypeGroup.dns)
    tld = Type("tld", TypeGroup.general)
    formatted_time = Type("formatted_time", TypeGroup.general)
    isotimestamp = Type("isotimestamp", TypeGroup.general)
    time = Type("time", TypeGroup.general)
    date = Type("date", TypeGroup.general)
    bytes = Type("bytes", TypeGroup.general, is_private=True)
    charset = Type("charset", TypeGroup.general)
    country_name = Type("country_name",
                        TypeGroup.general)  # XX not identifiable, user has to be told somehow there is such method
    phone = Type("phone", TypeGroup.general, "telephone number", ["telephone", "tel"], Checker.is_phone)

    unit = Type("unit", TypeGroup.general, "any physical quantity", [], Checker.is_unit)
    timestamp = Type("timestamp", TypeGroup.general, "time or date", ["time"], Checker.is_timestamp)
    ip = Type("ip", TypeGroup.general, "valid IP address", ["ip", "ipaddress"], Checker.is_ip)
    source_ip = Type("source_ip", TypeGroup.general, "valid source IP address",
                     ["sourceipaddress", "source", "src"], Checker.is_ip, usual_must_match=True)
    destination_ip = Type("destination_ip", TypeGroup.general, "valid destination IP address",
                          ["destinationipaddress", "destination", "dest", "dst"], Checker.is_ip, usual_must_match=True)
    port = Type("port", TypeGroup.general, "port", ["port", "prt"], lambda x: re.match(r"\d{1,5}", x),
                usual_must_match=True)
    cidr = Type("cidr", TypeGroup.general, "CIDR 127.0.0.1/32", ["cidr"], Checker.check_cidr)
    port_ip = Type("port_ip", TypeGroup.general, "IP in the form 1.2.3.4.port", [], reIpWithPort.match)
    any_ip = Type("any_ip", TypeGroup.general, "IP in the form 'any text 1.2.3.4 any text'", [],
                  lambda x: reAnyIp.search(x) and not Checker.is_ip(x))
    hostname = Type("hostname", TypeGroup.general, "2nd or 3rd domain name", ["fqdn", "hostname", "domain"],
                    reFqdn.match)
    email = Type("email", TypeGroup.general, "E-mail address", ["mail"],
                 lambda e: validate_email(e, check_dns=False, check_smtp=False))
    url = Type("url", TypeGroup.general, "URL starting with http/https", ["url", "uri", "location"],
               lambda s: reUrl.match(s) and "[.]" not in s)  # input "example[.]com" would be admitted as a valid URL)
    asn = Type("asn", TypeGroup.whois, "Autonomous system number", ["as", "asn", "asnumber"],
               lambda x: re.search(r'AS\d+', x) is not None)
    base64 = Type("base64", TypeGroup.general, "Text encoded with Base64", ["base64"], Checker.is_base64)
    quoted_printable = Type("quoted_printable", TypeGroup.general, "Text encoded as quotedprintable", [],
                            Checker.is_quopri)
    urlencode = Type("urlencode", TypeGroup.general, "Text encoded with urlencode", ["urlencode"], Checker.is_urlencode)
    wrong_url = Type("wrong_url", TypeGroup.general, "Deactivated URL", [], Checker.check_wrong_url)
    plaintext = Type("plaintext", TypeGroup.general, "Plain text", ["plaintext", "text"], lambda x: False)
    path = Type("path", usual_names=["path"], identify_method=lambda x: False)

    @staticmethod
    def get_computable_types(ignore_custom=False):
        """ List of all suitable fields that we may compute from a suitable output
        :type ignore_custom: bool Ignore types that may not be reachable because user input would be needed.
            These are TypeGroup.custom and instances of PickBase.
        """
        s = set()
        for (_, target_type), m in methods.items():
            if ignore_custom and (target_type.group == TypeGroup.custom or (isinstance(m, PickBase) and not m.default)):
                continue
            if target_type.is_private or target_type.is_disabled:
                continue
            s.add(target_type)
        return sorted(s)

    @staticmethod
    def get_guessable_types() -> List[Type]:
        """ these field types can be guessed from a string """
        return sorted([t for t in types if not t.is_disabled and (t.identify_method or t.usual_names)])

    @staticmethod
    def get_uml(flags):
        """ Return DOT UML source code of types and methods
            dashed nodes = identifiable automatically
            dashed edges = lambda-True relation (see _get_methods() help)
            squares = category border
            edge label = possible ways of processing (@PickBase decorated methods)

          FLAGS:
             * +1 to gray out disabled fields/methods
             * +2 to include usual field names
             * +16 wide-screen suitable (presenting) else tall graph
        """
        l = ['digraph { ',
             # 'rankdir=LR',  # XX I really wish to have TB, with rows: IP, plaintext, timestamp
             'label="Convey field types (dashed nodes = identifiable automatically, dashed edges = field identity)"']
        used = set()
        formatting_disabled = "[color=lightgray fontcolor=lightgray]" if flags & 1 else ""
        loop = []
        existing_edges = set()

        def add(start, target, disable_s):
            if (start, target) not in used:
                used.add((start, target))
                used.add((target, start))
                if not disable_s:
                    used.update([start, target])
                l.append(f"{start} -> {target}{disable_s}")
                if (target, start) in existing_edges:
                    l[-1] += "[dir=both]"
                return True
            return False

        for enabled, ((start, target), m) in [(True, f) for f in methods.items()] + [(False, f) for f in
                                                                                     methods_deleted.items()]:
            loop.append((enabled, start, target, m))
            existing_edges.add((start, target))
        # for enabled, ((start, target), m) in [(True, f) for f in methods.items()] + [(False, f) for f in methods_deleted.items()]:
        for enabled, start, target, m in loop:
            disable_s = "" if enabled else formatting_disabled
            if start.group != target.group and target.group.name != target and target.group in [TypeGroup.dns,
                                                                                                TypeGroup.nmap,
                                                                                                TypeGroup.custom]:
                # Every type that goes to ex: `dns`, continues to all `dns` subtypes. We want `hostname -> spf` to go through `dns`.
                # This is not the case of the group `whois` - this group has explicitly stated the path in methods,
                #   ex: `ip → whois → country` but also `phone → country`.
                add(start, target.group.name, disable_s)
                start = target.group.name
            if add(start, target, disable_s):
                if m is True:
                    l[-1] += "[style=dashed]"
                if isinstance(m, PickMethod):
                    l[-1] += '[label="choose ' + ",".join([name for name, _ in m.get_options()]) + '" fontsize=10]'
                elif isinstance(m, PickInput):
                    l[-1] += f'[label="{m.parameter_name}" fontsize=10]'
        for tg in TypeGroup:
            if tg is not TypeGroup.general:
                if tg.name in used:
                    l.append(f"{tg.name} [shape=box]")
                else:
                    l.append(f"{tg.name} [shape=box]{formatting_disabled}")
        for f in types:
            disable_s = "" if f in used else formatting_disabled
            label = [f.name]
            s = ""
            if f.description:
                label.append(f.description)
            if flags & 2:
                if f.usual_names:
                    label.append("usual names: " + ", ".join(f.usual_names))
                s = ' [label="' + r"\n".join(label) + '"]'
            l.append(f'{f.name}{s}{disable_s}')
            if f in Types.get_guessable_types():
                l.append(f'{f.name} [style=dashed]')
            # if f.is_private:
            #    l.append(f'{f.name} [shape=box]')

        # formatting mark to put one cluster below another, not aside
        if flags & 16:  # looks nicer in a presentation
            l.extend([
                # "spf -> timestamp[style=invis]",
                # "formatted_time -> plaintext[style=invis]",
            ])
        else:  # looks nicer in README.md
            l.extend([
                "port -> urlencode[style=invis]",
                "charset -> timestamp[style=invis]"
            ])
        l.append("}")
        return "\n".join(l)

    @staticmethod
    def _get_methods():
        """  These are known methods to compute a field from another field.
            They should return scalar or list.

            Method ~ lambda: Will be processed when converting one type to another
            Method ~ None: Will be skipped. (TypeGroup.custom fields usually have None.)
            Method ~ True: The user should not be offered to compute from the field to another from a longer distance.
                However if they already got the first type, it is the same as if they had the other.
                Example: (source_ip → ip) hostname from source_ip will be computed as if from an ip,
                Example: (abusemail → email) We do not want to offer the conversion hostname → whois → abusemail → email → hostname,
                    in other words we do not offer to compute a hostname from a hostname, however when selecting
                    an existing abusemail column, we offer conversion abusemail → (invisible email) → hostname
        """

        t = Types
        return {
            (t.source_ip, t.ip): True,
            (t.destination_ip, t.ip): True,
            (t.any_ip, t.ip): any_ip_ip,  # "91.222.204.175 93.171.205.34" -> "91.222.204.175" OR '"1.2.3.4"' -> 1.2.3.4
            (t.port_ip, t.ip): port_ip_ip,  # IP written with a port 91.222.204.175.23 -> 91.222.204.175
            (t.port_ip, t.port): port_ip_port,
            (t.url, t.hostname): url_hostname,
            (t.url, t.port): url_port,
            (t.hostname, t.ip): Checker.hostname_ips if Config.get("multiple_hostname_ip",
                                                                   "FIELDS") else Checker.hostname_ip,
            # (t.url, t.ip): Whois.url2ip,
            (t.ip, t.whois): Whois,
            (t.hostname, t.whoisdomain): lambda x: Whois(ip=None, hostname=x),
            # (t.asn, t.whois): Whois, # XX can be easily allowed, however Whois object will huff there is no IP prefix range
            (t.cidr, t.ip): Checker.cidr_ips if Config.get("multiple_cidr_ip", "FIELDS") else
            lambda x: str(ipaddress.ip_interface(x).ip),
            (t.whois, t.prefix): lambda x: str(x.get[0]),
            (t.whois, t.asn): lambda x: x.get[3],
            (t.whois, t.abusemail): lambda x: x.get[6],
            (t.whoisdomain, t.registrar_abusemail): lambda x: x.get[6],
            (t.whois, t.country): lambda x: x.get[5],
            (t.whois, t.netname): lambda x: x.get[4],
            (t.whois, t.csirt_contact):
                lambda x: Contacts.country2mail[x.get[5]] if x.get[5] in Contacts.country2mail else "",
            (t.whois, t.incident_contact): lambda x: x.get[2],
            (t.abusemail, t.cc_contact): lambda x: "; ".join(
                Contacts.mail2cc[domain] for domain in Contacts.get_domains(x) if domain in Contacts.mail2cc
            ),
            (t.plaintext, t.bytes): lambda x: x.encode("UTF-8"),
            (t.bytes, t.plaintext): Checker.bytes_plaintext,
            (t.bytes, t.base64): lambda x: b64encode(x).decode("UTF-8"),
            (t.base64, t.bytes): b64decode,
            (t.bytes, t.quoted_printable): lambda x: encodestring(x).decode("UTF-8"),
            (t.quoted_printable, t.bytes): decodestring,
            (t.urlencode, t.plaintext): lambda x: unquote(x),
            (t.plaintext, t.urlencode): lambda x: quote(x),
            (t.plaintext, t.external): None,
            (t.plaintext, t.code): None,
            (t.plaintext, t.reg): None,
            (t.bytes, t.charset): Checker.decode,
            (t.reg, t.reg_s): None,
            (t.reg, t.reg_m): None,
            (t.wrong_url, t.url): wrong_url_2_url,
            (t.hostname, t.url): lambda x: "http://" + x,
            # X (t.ip, t.url): lambda x: "http://" + x,
            (t.ip, t.hostname): lambda x: socket.gethostbyaddr(x)[0],
            (t.url, t.web): Web,
            (t.web, t.http_status): lambda x: x.get[0],
            (t.web, t.text): lambda x: x.get[1],
            (t.web, t.html): lambda x: x.get[2],
            (t.web, t.redirects): lambda x: x.get[3],
            (t.web, t.x_frame_options): lambda x: x.get[4],
            (t.web, t.csp): lambda x: x.get[5],
            (t.web, t.form_names): lambda x: x.get[6],
            (t.hostname, t.ports): nmap,
            (t.ip, t.ports): nmap,
            (t.ports, t.port): True,
            (t.hostname, t.spf): dig("SPF"),
            (t.hostname, t.txt): dig("TXT"),
            (t.hostname, t.a): dig("A"),
            (t.hostname, t.aaaa): dig("AAAA"),
            (t.hostname, t.ns): dig("NS"),
            (t.hostname, t.mx): dig("MX"),
            (t.hostname, t.dmarc): dig("DMARC"),
            (t.abusemail, t.email): True,
            (t.email, t.hostname): lambda x: x[x.index("@") + 1:],
            # (t.email, check legit mailbox)
            (t.country_name, t.country): address_country,
            (t.country, t.country_name): lambda x: country_codes[x],
            (t.country, t.csirt_contact): lambda x: Contacts.country2mail[x] if x in Contacts.country2mail else "",
            (t.phone, t.country): phone_country,
            (t.hostname, t.tld): Checker.HostnameTld,
            (t.prefix, t.cidr): Checker.prefix_cidr,
            (t.redirects, t.url): True,
            (t.tld, t.country): True,
            (t.timestamp, t.formatted_time): Checker.time_format,
            (t.timestamp, t.isotimestamp): Checker.isotimestamp,
            (t.timestamp, t.date): Checker.date,
            (t.timestamp, t.time): lambda val: Checker.parse_timestamp(val).time(),
            (t.unit, t.plaintext): Checker.unit_expand,
            (t.plaintext, t.unit): Checker.unit_expand
        }


def get_module_from_path(path):
    if not Path(path).is_file():
        return False
    spec = importlib.util.spec_from_file_location("", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



import ipaddress
import logging
import re
import socket
from base64 import b64decode
from datetime import datetime
from quopri import decodestring
from urllib.parse import unquote, urlparse

import dateutil.parser
from chardet import detect
from netaddr import IPNetwork, IPRange
from pint import UnitRegistry

from .convert import reFqdn, reUrl, wrong_url_2_url
from .decorators import PickInput, PickMethod
from .infodicts import phone_regex_match
from .utils import timeout

logger = logging.getLogger(__name__)
pint = UnitRegistry()


class Checker:
    """ To not pollute the namespace, we put the methods here """

    hostname_ips_cache = {}
    hostname_cache = {}

    @staticmethod
    def hostname_ips(val):
        if val not in Checker.hostname_ips_cache:
            try:
                Checker.hostname_ips_cache[val] = list(
                    {addr[4][0] for addr in timeout(15, socket.getaddrinfo, val, None)})
            except (TimeoutError, OSError, ValueError) as e:
                Checker.hostname_ips_cache[val] = []
        return Checker.hostname_ips_cache[val]

    @classmethod
    def hostname_ip(cls, val):
        if val not in cls.hostname_cache:
            try:
                cls.hostname_cache[val] = timeout(3, socket.gethostbyname, val)
            except (TimeoutError, OSError, ValueError) as e:
                logger.warning(f"Hostname {val}: {e}")
                cls.hostname_cache[val] = []
        return cls.hostname_cache[val]

    @staticmethod
    def check_cidr(cidr):
        try:
            ipaddress.ip_interface(cidr)
            try:
                ipaddress.ip_address(cidr)
            except ValueError:  # "1.2.3.4" fail, "1.2.3.4/24" pass
                return True
        except ValueError:
            return False

    @staticmethod
    def is_base64(x):
        """ We prefer as base64-encoded only such strings that could be decoded to UTF-8.
            Because otherwise nearly any ASCII input with the correct padding
            would be considered as base64 (ex: port number), even when well readable at first sight.
            If not UTF-8 decodable, we check the minimal length and penalize.
        """
        if not re.match(r"^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$", x):
            return False

        try:
            s = b64decode(x).decode("UTF-8")
            if s:
                if len(re.sub(r"[^a-zA-Z0-9 ]", "", s)) / len(s) > 0.9:
                    # Majority of characters in the decoded string must be alphanumerical
                    # in order we think this might be a base64-encoded string.
                    # Imagine string 'ahojahoj' which is base64-decodable but gibberish.
                    return True
        except (UnicodeDecodeError, ValueError):
            try:
                if len(x) > 10 and b64decode(x):
                    # if the string is long enough, we admit it can be encoded in another charset
                    return -0.5
            except (UnicodeDecodeError, ValueError):
                pass

    @staticmethod
    def is_ip(ip):
        """ True, if IP is well formatted IPv4 or IPv6 """
        try:
            ipaddress.ip_address(ip)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_quopri(x):
        try:
            # When 'rmx1u2916Gv9IGv58iBw7Gwg7+FiZWxza+kg82R5Lg==' is decoded, only last '=' is stripped out.
            decoded = decodestring(x).decode("UTF-8").rstrip("=")
            return decoded and decoded != x.rstrip("=")
        except (UnicodeDecodeError, ValueError):
            return False

    @staticmethod
    def bytes_plaintext(x):
        try:
            return x.decode("UTF-8").replace("\n", r"\n")
        except UnicodeDecodeError:
            enc = detect(x)["encoding"]
            # Czech phrase: "Žluťoučký kůň pěl ďábelské ódy.".encode("iso-8859-2")
            # results in turkish {'encoding': 'ISO-8859-9', 'confidence': 0.47567063613812527, 'language': 'Turkish'}
            # which is not perfect but better than nothing.
            if enc:
                return x.decode(enc)
            return x

    @staticmethod
    def is_urlencode(x):
        return unquote(x) != x

    @staticmethod
    def check_wrong_url(wrong):
        if (reUrl.match(wrong) and "[.]" not in wrong) or reFqdn.match(wrong) or Checker.is_ip(wrong):
            # input "example[.]com" would be admitted as a valid URL)
            return False
        s = wrong_url_2_url(wrong, make=False)
        s2 = wrong_url_2_url(wrong, make=True)
        try:
            netloc = urlparse(s2).netloc
        except ValueError:
            # ex: s2='http://[07/Feb/2021:00:32:43 -0500] "GET /wp-json/wp/v2/posts HTTP/1.1" 200 1912'
            return False
        # Xm = reUrl.match(s2)
        # X(m and m.span(0)[1] == len(s2) and "." in s2):
        if reFqdn.match(s) or reFqdn.match(netloc) or Checker.is_ip(netloc):
            # Xwe impose full match
            # Xstring created from "x"*100 would be submitted for a valid URL
            return True
        return False

    @staticmethod
    @PickInput
    def decode(val, encoding="utf-8"):
        if type(val) is str:
            val = val.encode("utf-8")
        return val.decode(encoding).replace("\n", r"\n")

    @staticmethod
    @PickInput
    def time_format(val, fmt):
        return dateutil.parser.parse(val, fuzzy=True).strftime(fmt)

    # noinspection PyBroadException
    @staticmethod
    def is_unit(val):
        if len(val) == 1:
            # we want prevent "a" being considered a time unit ("a" == "1 year" according to pint 0.16.1)
            return False
        try:
            e = pint.parse_expression(val)
            return e and pint.get_compatible_units(e)
        except Exception:
            return False

    @staticmethod
    @PickInput
    def unit_expand(val, unit=None):
        try:
            e = pint.parse_expression(val)
            if unit:
                return str(e.to(unit))
            else:
                l = []
                for unit in pint.get_compatible_units(e):
                    l.append(str(e.to(unit)))
                return l
        except (ValueError, AttributeError):
            return ""

    @classmethod
    def is_phone(cls, val):
        return phone_regex_match(val) and not cls.is_timestamp(val)

    @staticmethod
    def is_timestamp(val):
        reference = datetime.now()

        # identify UNIX time
        try:
            o = datetime.fromtimestamp(float(val))
        except ValueError:
            pass  # val is not a number, try different methods
        else:
            # consider only nowadays timestamp a timestamp
            # prevent "12345" to be considered as a timestamp from the beginning of the Unix epoch (1970)
            return reference.year - 5 < o.year < reference.year + 1

        # identify straight textual representation (ex: 2021-02-12), easily parsable by not-fuzzy `dateutil`
        try:
            o = dateutil.parser.parse(val)
            if not (o.second == o.minute == o.hour == 0 and o.day == reference.day and o.month == reference.month):
                # if parser accepts a plain number, ex: 1920,
                # it thinks this is a year without time (assigns midnight) and without date (assigns current date)
                # -> 1920-05-05 00:00 (on 5th May) – we try to skip such result.
                # If does not match, it seems this is a valid timestamp and we hereby return True.
                return True
        except ValueError:
            pass  # val is a random textual representation or a UNIX time
        except OverflowError:
            return False

        # identify random textual representation of a timestamp
        try:
            o = Checker.parse_timestamp(val)
            if 2100 > o.year > 1900 and not (o.second == o.minute == o.hour == 0 and o.day == o.month == 1):
                # this year seems reasonable and it is not suspicious
                # fuzzy search often return crazy things like year from port number - these records have 1 Jan midnight
                return True
        except (ValueError, TypeError):
            pass
        return False

    default_datetime = datetime(1, 1, 1)

    @staticmethod
    def parse_timestamp(val):
        # UNIX time
        try:
            return datetime.fromtimestamp(float(val))
        except ValueError:
            pass  # val is not a number

        # random textual representation of a timestamp
        try:
            return dateutil.parser.parse(val, fuzzy=True, default=Checker.default_datetime)
        except OverflowError:
            return Checker.default_datetime

    @staticmethod
    def isotimestamp(val):
        o = Checker.parse_timestamp(val)
        s = o.isoformat()
        if o.year == 1:  # this is a fake year from our default_datetime object
            s = "0000-00-00" + s[10:]
        return s

    @staticmethod
    def date(val):
        o = Checker.parse_timestamp(val)
        if o.year == 1:  # this is a fake year from our default_datetime object
            return ""
        return o.date()

    @staticmethod
    @PickMethod("all")
    class HostnameTld:
        @staticmethod
        def all(x):
            """ take all TLD """
            x = x[x.rindex(".") + 1:]
            if not x.isdigit():
                return x

        @classmethod
        def ccTLD(cls, x):
            """ country code only """
            x = cls.all(x)
            return x if len(x) == 2 else ""

        @classmethod
        def gTLD(cls, x):
            """ generic only """
            x = cls.all(x)
            return x if len(x) != 2 else ""

    @staticmethod
    def prefix_cidr(val):
        if "/" in val:
            # currently, we do not guarantee prefix will be in the form "... - ...",
            # it may has directly CIDR form if WHOIS tells us so
            return val
        return IPRange(*val.split("-")).cidrs()

    @staticmethod
    def cidr_ips(val):
        return [ip for ip in IPNetwork(val)]



import logging
import re
from urllib.parse import urlsplit

from .config import Config
from .utils import print_atomic
import subprocess

logger = logging.getLogger(__name__)

reIpWithPort = re.compile(r"((\d{1,3}\.){4})(\d+)")
reAnyIp = re.compile(r"\"?((\d{1,3}\.){3}(\d{1,3}))")
reFqdn = re.compile(
    r"(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-_]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)")  # Xtoo long, infinite loop: ^(((([A-Za-z0-9]+){1,63}\.)|(([A-Za-z0-9]+(\-)+[A-Za-z0-9]+){1,63}\.))+){1,255}$
reUrl = re.compile(r'[htps]*://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
# reBase64 = re.compile('^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$')


def wrong_url_2_url(s, make=True):
    s = re.sub("^hxxp", 'http', s, flags=re.I).replace("[.]", ".").replace("(.)", ".").replace("[:]", ":")
    if make and not s.lower().startswith("http"):
        s = "http://" + s
    return s


def any_ip_ip(s):
    m = reAnyIp.search(s)
    if m:
        return m.group(1)


def port_ip_ip(s):
    m = reIpWithPort.match(s)
    if m:
        return m.group(1).rstrip(".")


def port_ip_port(s):
    m = reIpWithPort.match(s)
    if m:
        return m.group(3)


def url_port(s):
    s = s.split(":")[1]
    return re.match(r"^(\d*)", s).group(1)


def url_hostname(url):
    """ Covers both use cases "http://example.com..." and "example.com..." """
    s = urlsplit(url)
    s = s.netloc or s.path.split("/")[:1][0]
    return s.split(":")[0]  # strips port

def dig(rr):
    def dig_query(query):
        print_atomic(f"Digging {rr} of {query}")
        if rr == "SPF":
            t = "TXT"
        elif rr == "DMARC":
            query = "_dmarc." + query
            t = "TXT"
        else:
            t = rr
        try:
            text = subprocess.check_output(["dig", "+short", "-t", t, query, "+timeout=1"]).decode("utf-8")
        except FileNotFoundError:
            Config.missing_dependency("dnsutils")
        if text.startswith(";;"):
            return None
        spl = text.split("\n")[:-1]
        if t == "TXT":
            spl = [r[1:-1] for r in spl if r.startswith('"') and r.endswith('"')]  # row without " may be CNAME redirect
            if rr == "SPF":
                return [r for r in spl if r.startswith('v=spf')]
            elif rr == "TXT":
                return [r for r in spl if not r.startswith('v=spf')]
        logger.debug(f"Dug {spl}")
        return spl

    return dig_query


# @PickInput XX not yet possible to PickInput
# 1. since for port '80' wizzard loads the port '8' and then '80'
# 2. since we cannot specify `--field ports[80,443]` because comma ',' is taken for a delmiiter between FIELD and COLUMN
#   and pair quoting char '[]' is not allowed in csv.reader that parses CLI
def nmap(val, port=""):
    """
    :type port: int Port to scan, you may delimit by a comma. Ex: `80, 443`
    """
    logger.info(f"NMAPing {val}...")
    try:
        cmd = ["nmap", val]
        if port:
            cmd.extend(["-p", port])
        text = subprocess.run(cmd, stdout=subprocess.PIPE).stdout.decode("utf-8")
    except FileNotFoundError:
        Config.missing_dependency("nmap")
    text = text[text.find("PORT"):]
    text = text[text.find("\n") + 1:]
    text = text[:text.find("\n\n")]
    if Config.get("multiple_nmap_ports", "FIELDS"):
        l = []
        for row in text.split("\n"):
            l.append(int(re.match(r"(\d+)", row).group(1)))
        return l

    return text
